#ifdef FHGFS_OPENTK_IBVERBS

#include "IBVSocket.h"


#define IBVSOCKET_CONN_TIMEOUT_MS      3000
#define IBVSOCKET_LISTEN_BACKLOG       32


void IBVSocket_init(IBVSocket* _this)
{
   memset(_this, 0, sizeof(*_this) );

   _this->sockValid = false;

   _this->cm_channel = rdma_create_event_channel();
   if(!_this->cm_channel)
   {
      fprintf(stderr, "%d:%s: rdma_create_event_channel failed\n", __LINE__, __func__);
      return;
   }

   if(rdma_create_id(_this->cm_channel, &_this->cm_id, NULL, RDMA_PS_TCP) )
   {
      fprintf(stderr, "%d:%s: rdma_create_id failed\n", __LINE__, __func__);
      return;
   }

   _this->sockValid = true;

   return;
}

/**
 * @param commContext belongs to this object (so do not use or free it after calling this!)
 */
void __IBVSocket_initFromCommContext(IBVSocket* _this, struct rdma_cm_id* cm_id,
   IBVCommContext* commContext)
{
   memset(_this, 0, sizeof(*_this) );

   _this->cm_id = cm_id;
   _this->commContext = commContext;

   _this->sockValid = true;

   return;
}

IBVSocket* IBVSocket_construct()
{
   IBVSocket* _this = (IBVSocket*)malloc(sizeof(*_this) );

   IBVSocket_init(_this);

   return _this;
}

IBVSocket* __IBVSocket_constructFromCommContext(struct rdma_cm_id* cm_id,
   IBVCommContext* commContext)
{
   IBVSocket* _this = (IBVSocket*)malloc(sizeof(*_this) );

   __IBVSocket_initFromCommContext(_this, cm_id, commContext);

   return _this;
}

void IBVSocket_uninit(IBVSocket* _this)
{
   __IBVSocket_close(_this);
}

void IBVSocket_destruct(IBVSocket* _this)
{
   IBVSocket_uninit(_this);

   free(_this);
}

bool IBVSocket_rdmaDevicesExist()
{
   bool devicesExist;

   int numDevices = 1;
   struct ibv_context** devicesRes;

   devicesRes = rdma_get_devices(&numDevices);

   devicesExist = (devicesRes != NULL);

   rdma_free_devices(devicesRes);

   return devicesExist;
}

/**
 * Prepare ibverbs for multi-threading.
 * Call this only once in your program.
 *
 * Note: There is no corresponding uninit-method that needs to be called.
 */
void IBVSocket_fork_init_once()
{
   ibv_fork_init();
}

bool IBVSocket_connectByName(IBVSocket* _this, const char* hostname, unsigned short port,
   IBVCommConfig* commCfg)
{
   struct addrinfo *res;
   struct addrinfo hints;

   int getInfoRes;
   struct in_addr ipaddress;

   memset(&hints, 0, sizeof(hints) );
   hints.ai_family = AF_UNSPEC;
   hints.ai_socktype = SOCK_STREAM;

   getInfoRes = getaddrinfo(hostname, NULL, &hints, &res);

   if(getInfoRes < 0)
   {
      fprintf(stderr, "%d:%s: %s for %s:%d\n", __LINE__, __func__, gai_strerror(getInfoRes),
         hostname, (int)port);

      return false;
   }

   ipaddress.s_addr = ( (struct sockaddr_in*)res->ai_addr)->sin_addr.s_addr;


   // clean-up
   freeaddrinfo(res);


   return IBVSocket_connectByIP(_this, &ipaddress, port, commCfg);
}

bool IBVSocket_connectByIP(IBVSocket* _this, struct in_addr* ipaddress, unsigned short port,
   IBVCommConfig* commCfg)
{
   struct rdma_cm_event* event;
   struct sockaddr_in sin;
   bool createContextRes;
   struct rdma_conn_param conn_param;
   bool parseCommDestRes;


   // resolve IP address...

   sin.sin_addr.s_addr = ipaddress->s_addr;
   sin.sin_family = AF_INET;
   sin.sin_port = htons(port);

   if(rdma_resolve_addr(_this->cm_id, NULL, (struct sockaddr*)&sin, IBVSOCKET_CONN_TIMEOUT_MS) )
   {
      fprintf(stderr, "%d:%s: rdma_resolve_addr failed\n",
         __LINE__, __func__);
      goto err_invalidateSock;
   }

   if(rdma_get_cm_event(_this->cm_channel, &event))
      goto err_invalidateSock;

   if(event->event != RDMA_CM_EVENT_ADDR_RESOLVED)
   {
      fprintf(stderr, "%d:%s:%d: unexpected CM event %d: %s\n",
         __LINE__, __func__, __LINE__, event->event, rdma_event_str(event->event) );
      goto err_ack_and_invalidateSock;
   }

   rdma_ack_cm_event(event);

   // resolve route...

   if(rdma_resolve_route(_this->cm_id, IBVSOCKET_CONN_TIMEOUT_MS) )
   {
      fprintf(stderr, "%d:%s: rdma_resolve_route failed\n",
         __LINE__, __func__);
      goto err_invalidateSock;
   }

   if(rdma_get_cm_event(_this->cm_channel, &event))
      goto err_invalidateSock;

   if(event->event != RDMA_CM_EVENT_ROUTE_RESOLVED)
   {
      fprintf(stderr, "%d:%s:%d: unexpected CM event %d: %s\n",
         __LINE__, __func__, __LINE__, event->event, rdma_event_str(event->event) );
      goto err_ack_and_invalidateSock;
   }

   rdma_ack_cm_event(event);

   // create comm context...

   createContextRes = __IBVSocket_createCommContext(_this, _this->cm_id, commCfg,
      &_this->commContext);
   if(!createContextRes)
   {
      fprintf(stderr, "%d:%s: creation of CommContext failed\n", __LINE__, __func__);
      goto err_invalidateSock;
   }

   // establish connection...

   __IBVSocket_initCommDest(_this->commContext, &_this->localDest);

   memset(&conn_param, 0, sizeof(conn_param) );
   conn_param.responder_resources = 1;
   conn_param.initiator_depth = 1;
   conn_param.flow_control = 1;
   conn_param.retry_count = 7; // (3 bits)
   conn_param.rnr_retry_count = 7; // rnr = receiver not ready (3 bits, 7 means infinity)
   conn_param.private_data = &_this->localDest;
   conn_param.private_data_len = sizeof(_this->localDest);

   if(rdma_connect(_this->cm_id, &conn_param))
   {
      fprintf(stderr, "%d:%s: rdma_connect failure\n", __LINE__, __func__);
      goto err_invalidateSock;
   }

   if(rdma_get_cm_event(_this->cm_channel, &event))
      goto err_invalidateSock;

   if(event->event != RDMA_CM_EVENT_ESTABLISHED)
   {
      fprintf(stderr, "%d:%s:%d: unexpected CM event %d: %s\n",
         __LINE__, __func__, __LINE__, event->event, rdma_event_str(event->event) );
      goto err_ack_and_invalidateSock;
   }

   parseCommDestRes = __IBVSocket_parseCommDest(
      event->param.conn.private_data, event->param.conn.private_data_len, &_this->remoteDest);
   if(!parseCommDestRes)
   {
      fprintf(stderr, "%d:%s: bad private data received. len: %d\n",
         __LINE__, __func__, event->param.conn.private_data_len);
      goto err_ack_and_invalidateSock;
   }

   rdma_ack_cm_event(event);

   return true;


err_ack_and_invalidateSock:
   rdma_ack_cm_event(event);
err_invalidateSock:
   _this->errState = -1;

   return false;
}

/**
 * @return true on success
 */
bool IBVSocket_bind(IBVSocket* _this, unsigned short port)
{
   in_addr_t ipAddr = INADDR_ANY;

   return IBVSocket_bindToAddr(_this, ipAddr, port);
}

bool IBVSocket_bindToAddr(IBVSocket* _this, in_addr_t ipAddr, unsigned short port)
{
   struct sockaddr_in bindAddr;

   bindAddr.sin_family = AF_INET;
   bindAddr.sin_addr.s_addr = ipAddr;
   bindAddr.sin_port = htons(port);

   if(rdma_bind_addr(_this->cm_id, (struct sockaddr*)&bindAddr) )
   {
      //fprintf(stderr, "%s:%d rdma_bind_addr failed (port: %d)\n", __func__, __LINE__,
      //   (int)port); // debug in
      goto err_invalidateSock;
   }

   return true;


err_invalidateSock:
   _this->errState = -1;

   return false;
}

/**
 * Note: This also inits the delayedCmEventsQueue.
 *
 * @return true on success
 */
bool IBVSocket_listen(IBVSocket* _this)
{
   if(rdma_listen(_this->cm_id, IBVSOCKET_LISTEN_BACKLOG) )
   {
      fprintf(stderr, "%d:%s: rdma_listen failed\n", __LINE__, __func__);
      goto err_invalidateSock;
   }

   // init delayed events queue
   _this->delayedCmEventsQ = new CmEventQueue();


   return true;


err_invalidateSock:
   _this->errState = -1;

   return false;
}

/**
 * Note: Call IBVSocket_checkDelayedEvents() after this to find out whether more events
 *    are waiting.
 * Note: Because of the special way ibverbs accept connections, it is possible that we receive
 *    some other events here as well (e.g. a child socket disconnect). In these cases,
 *    ACCEPTRES_IGNORE will be returned.
 *
 * @param outAcceptedSock only valid when ACCEPTRES_SUCCESS is returned
 * @param peerAddr (out) peer address
 * @param peerAddrLen (out) length of peer address
 * @return ACCEPTRES_IGNORE in case an irrelevant event occured
 */
IBVSocket_AcceptRes IBVSocket_accept(IBVSocket* _this, IBVSocket** outAcceptedSock,
   struct sockaddr* peerAddr, socklen_t* peerAddrLen)
{
   struct rdma_cm_event* event = NULL;
   IBVCommContext* childCommContext = NULL;
   IBVSocket* acceptedSock = NULL; // auto-destructed on error/ignore
   IBVCommDest* childRemoteDest = NULL; // auto-freed on error/ignore

   *outAcceptedSock = NULL;


   // get next waiting event from delay-queue or from event channel

   if(_this->delayedCmEventsQ->size() )
   {
      event = _this->delayedCmEventsQ->front();
      _this->delayedCmEventsQ->pop();
   }
   else
   if(rdma_get_cm_event(_this->cm_channel, &event) )
   {
      _this->errState = -1;
      return ACCEPTRES_ERR;
   }


   // handle event type

   switch(event->event)
   {
      case RDMA_CM_EVENT_CONNECT_REQUEST:
      {
         // got an incoming 'connect request' => check validity of private data and accept/reject

         bool createContextRes;
         struct rdma_conn_param conn_param;
         bool parseCommDestRes;
         IBVCommConfig commCfg;

         struct rdma_cm_id* child_cm_id = (struct rdma_cm_id*)event->id;

         //*peerAddrLen = sizeof(struct sockaddr_in);
         //memcpy(peerAddr, &child_cm_id->route.addr.dst_addr, *peerAddrLen);


         // parse private data to get remote dest

         parseCommDestRes = __IBVSocket_parseCommDest(
            event->param.conn.private_data, event->param.conn.private_data_len, &childRemoteDest);
         if(!parseCommDestRes)
         { // bad private data => reject connection

            fprintf(stderr, "%d:%s: bad private data received. len: %d\n",
               __LINE__, __func__, event->param.conn.private_data_len);

            if(rdma_reject(child_cm_id, NULL, 0) )
               fprintf(stderr, "%d:%s: rdma_reject failed\n", __LINE__, __func__);

            goto ignore;
         }


         // private data (remote dest) okay => create local comm context and socket instance

         // (we use the buffer config as suggested by the connecting peer)
         commCfg.bufNum = childRemoteDest->recvBufNum;
         commCfg.bufSize = childRemoteDest->recvBufSize;

         createContextRes = __IBVSocket_createCommContext(_this, child_cm_id, &commCfg,
            &childCommContext);
         if(!createContextRes)
         {
            fprintf(stderr, "%d:%s: creation of CommContext failed\n", __LINE__, __func__);

            if(rdma_reject(child_cm_id, NULL, 0) )
               fprintf(stderr, "%d:%s: rdma_reject failed\n", __LINE__, __func__);

            goto ignore;
         }

         acceptedSock = __IBVSocket_constructFromCommContext(
            child_cm_id, childCommContext);

         acceptedSock->remoteDest = childRemoteDest;
         childRemoteDest = NULL;  // would otherwise be destroyed at 'ignore'


         // send accept message (with local destination info)

         __IBVSocket_initCommDest(childCommContext, &acceptedSock->localDest);

         memset(&conn_param, 0, sizeof(conn_param) );
         conn_param.responder_resources = 1;
         conn_param.initiator_depth = 1;
         conn_param.flow_control = 1;
         conn_param.retry_count = 7; // (3 bits)
         conn_param.rnr_retry_count = 7; // rnr = receiver not ready (3 bits, 7 means infinity)
         conn_param.private_data = &acceptedSock->localDest;
         conn_param.private_data_len = sizeof(acceptedSock->localDest);

         if(rdma_accept(child_cm_id, &conn_param) )
         {
            fprintf(stderr, "%d:%s: rdma_accept failed\n", __LINE__, __func__);

            goto ignore;
         }

         child_cm_id->context = acceptedSock;
         acceptedSock = NULL; // would otherwise be destroyed at 'ignore'

         goto ignore;
      } break;

      case RDMA_CM_EVENT_ESTABLISHED:
      {
         // received 'established' (this is what we've actually been waiting for!)

         *peerAddrLen = sizeof(struct sockaddr_in);
         memcpy(peerAddr, &event->id->route.addr.dst_addr, *peerAddrLen);

         *outAcceptedSock = (IBVSocket*)event->id->context;

         rdma_ack_cm_event(event);

         return ACCEPTRES_SUCCESS;
      } break;

      case RDMA_CM_EVENT_DISCONNECTED:
      {
         //printf("%s:%d: Disconnect event %d: %s\n",
         //   __func__, __LINE__, event->event, rdma_event_str(event->event) ); // debug in

         rdma_cm_id* eventID = event->id;

         IBVSocket* childSock = (IBVSocket*)eventID->context;
         if(childSock)
         {
            // Note: Check error state, because in case of an error, the program is probably already
            // (and has the obligation to be) calling IBVSocket_close from another thread
            if(!childSock->errState)
               rdma_disconnect(eventID);
         }

         goto ignore;
      } break;

      case RDMA_CM_EVENT_CONNECT_ERROR:
      case RDMA_CM_EVENT_UNREACHABLE:
      {
         fprintf(stderr, "%d:%s: connect error event while waiting for 'established': %d (%s)\n",
            __LINE__, __func__, event->event, rdma_event_str(event->event) );

         acceptedSock = (IBVSocket*)event->id->context; // will be destroyed at 'ignore'

         goto ignore;
      } break;

      default:
      { // ignore other events
         printf("%s:%d: Ignoring conn manager event %d: %s\n",
            __func__, __LINE__, event->event, rdma_event_str(event->event) ); // debug in

         goto ignore;
      } break;
   }


   // irrelevant event (irrelevant for the caller)
ignore:
   rdma_ack_cm_event(event);

   SAFE_FREE(childRemoteDest);
   if(acceptedSock)
      IBVSocket_destruct(acceptedSock);

   *outAcceptedSock = NULL;

   return ACCEPTRES_IGNORE;
}

bool IBVSocket_shutdown(IBVSocket* _this)
{
   IBVCommContext* commContext = _this->commContext;

   if(_this->errState)
      return true; // true, because the conn is down anyways

   if(!commContext)
      return true; // this socket has never been connected

   if(commContext->incompleteSend.numAvailable)
   { // wait for all incomplete sends
      int waitRes;

      waitRes = __IBVSocket_waitForTotalSendCompletion(
         _this, commContext->incompleteSend.numAvailable, 0, 0);
      if(waitRes < 0)
      {
         fprintf(stderr, "%s: Waiting for incomplete send requests failed\n", __func__);
         return false;
      }
   }

   // __IBVSocket_disconnect(_this, false); // note: read the method comments before using this

   return true;
}

/**
 * Continues an incomplete former recv() by returning immediately available data from the
 * corresponding buffer.
 */
ssize_t __IBVSocket_recvContinueIncomplete(IBVSocket* _this, char* buf, size_t bufLen)
{
   IBVCommContext* commContext = _this->commContext;
   int completedOffset = commContext->incompleteRecv.completedOffset;
   size_t availableLen = commContext->incompleteRecv.wc.byte_len - completedOffset;
   size_t bufIndex = commContext->incompleteRecv.wc.wr_id - IBVSOCKET_RECV_WORK_ID_OFFSET;


   if(availableLen <= bufLen)
   { // old data fits completely into buf
      memcpy(buf, &(commContext->recvBufs)[bufIndex][completedOffset], availableLen);

      commContext->incompleteRecv.isAvailable = 0;

      if(__IBVSocket_postRecv(_this, _this->commContext, bufIndex) )
         goto err_invalidateSock;

      return availableLen;
   }
   else
   { // still too much data for the buf => copy partially
      memcpy(buf, &(commContext->recvBufs)[bufIndex][completedOffset], bufLen);

      commContext->incompleteRecv.completedOffset += bufLen;

      return bufLen;
   }


err_invalidateSock:
   _this->errState = -1;

   return -1;
}


ssize_t IBVSocket_recv(IBVSocket* _this, char* buf, size_t bufLen, int flags)
{
   const int timeoutMS = 1024*1024;
   ssize_t recvTRes;

   do
   {
      recvTRes = IBVSocket_recvT(_this, buf, bufLen, flags, timeoutMS);
   } while(!recvTRes);

   return recvTRes;
}


/**
 * @return number of received bytes on success, 0 on timeout, -1 on error
 */
ssize_t IBVSocket_recvT(IBVSocket* _this, char* buf, size_t bufLen, int flags, int timeoutMS)
{
   IBVCommContext* commContext = _this->commContext;
   struct ibv_wc* wc = &commContext->incompleteRecv.wc;
   int recvWCRes;
   size_t bufIndex;

   // check whether an old buffer has not been fully read yet
   if(commContext->incompleteRecv.isAvailable)
      return __IBVSocket_recvContinueIncomplete(_this, buf, bufLen);

   // recv new buffer
   recvWCRes = __IBVSocket_recvWC(_this, timeoutMS, wc);
   if(recvWCRes <= 0)
   {
      if(!recvWCRes)
         return 0; // timeout

      goto err_invalidateSock; // error occured
   }

   bufIndex = wc->wr_id - IBVSOCKET_RECV_WORK_ID_OFFSET;

   if(wc->byte_len <= bufLen)
   { // received data fits completely into buf
      memcpy(buf, commContext->recvBufs[bufIndex], wc->byte_len);

      //usleep(1); // debug in

      if(__IBVSocket_postRecv(_this, _this->commContext, bufIndex) )
         goto err_invalidateSock;

      return wc->byte_len;
   }
   else
   { // too much data for the buf => prepare to continue later
      memcpy(buf, commContext->recvBufs[bufIndex], bufLen);

      commContext->incompleteRecv.completedOffset = bufLen;
      commContext->incompleteRecv.isAvailable = 1;

      return bufLen;
   }


err_invalidateSock:
   _this->errState = -1;

   return -1;
}

ssize_t IBVSocket_send(IBVSocket* _this, const char* buf, size_t bufLen, int flags)
{
   IBVCommContext* commContext = _this->commContext;
   int postRes;
   size_t postedLen = 0;
   int currentPostLen;
   int waitRes;

   do
   {
      // note: we only poll for completed sends after we used up all (!) available bufs

      if(commContext->incompleteSend.numAvailable == commContext->commCfg.bufNum)
      { // wait for all (!) incomplete sends
         waitRes = __IBVSocket_waitForTotalSendCompletion(
            _this, commContext->incompleteSend.numAvailable, 0, 0);
         if(waitRes < 0)
            goto err_invalidateSock;

         commContext->incompleteSend.numAvailable = 0;
      }

      currentPostLen = FHGFS_MIN(bufLen-postedLen, commContext->commCfg.bufSize);

      // note: waitForUsedSendBufsReset can change incompleteSend.numAvailable, so it
      //    must be done before the memcpy()
      if(__IBVSocket_waitForUsedSendBufsReset(_this) )
         goto err_invalidateSock;

      memcpy( (commContext->sendBufs)[commContext->incompleteSend.numAvailable], &buf[postedLen],
         currentPostLen);

      commContext->numUsedSendBufs++;

      postRes = __IBVSocket_postSend(_this,
         commContext->incompleteSend.numAvailable, currentPostLen);
      if(postRes)
         goto err_invalidateSock;

      commContext->incompleteSend.numAvailable++;

      postedLen += currentPostLen;

   } while(postedLen < bufLen);

   return (ssize_t)bufLen;


err_invalidateSock:
   _this->errState = -1;

   return -1;
}


int __IBVSocket_registerBuf(IBVCommContext* commContext, void* buf, size_t bufLen,
   struct ibv_mr** outMR)
{
   // Note: IB spec says:
   //    "The consumer is not allowed to assign remote-write or remote-atomic to
   //    a memory region that has not been assigned local-write."
   enum ibv_access_flags accessFlags = (enum ibv_access_flags)
      (IBV_ACCESS_REMOTE_READ | IBV_ACCESS_REMOTE_WRITE | IBV_ACCESS_LOCAL_WRITE);

   *outMR = ibv_reg_mr(commContext->pd, buf, bufLen, accessFlags);
   if(!*outMR)
   {
      fprintf(stderr, "%d:%s: Couldn't allocate MR\n", __LINE__, __func__);
      return -1;
   }

   return 0;
}


char* __IBVSocket_allocAndRegisterBuf(IBVCommContext* commContext, size_t bufLen,
   struct ibv_mr** outMR)
{
   void* buf;
   int registerRes;

   int allocRes = posix_memalign(&buf, sysconf(_SC_PAGESIZE), bufLen);
   if(allocRes)
   {
      fprintf(stderr, "%d:%s: Couldn't allocate work buf.\n", __LINE__, __func__);
      return NULL;
   }

   memset(buf, 0, bufLen);

   registerRes = __IBVSocket_registerBuf(commContext, buf, bufLen, outMR);
   if(registerRes < 0)
   {
      free(buf);
      return NULL;
   }

   return (char*)buf;
}

bool __IBVSocket_createCommContext(IBVSocket* _this, struct rdma_cm_id* cm_id,
   IBVCommConfig* commCfg, IBVCommContext** outCommContext)
{
   IBVCommContext* commContext;
   int registerControlRes;
   int registerControlResReset;
   struct ibv_qp_init_attr qpInitAttr;
   unsigned i;

   commContext = (IBVCommContext*)malloc(sizeof(*commContext) );
   if(!commContext)
      goto err_cleanup;

   memset(commContext, 0, sizeof(*commContext) );

   commContext->context = cm_id->verbs;
   if(!commContext->context)
   {
      fprintf(stderr, "%d:%s: Unbound cm_id!!\n", __LINE__, __func__);
      goto err_cleanup;
   }

   commContext->pd = ibv_alloc_pd(commContext->context);
   if(!commContext->pd)
   {
      fprintf(stderr, "%d:%s: Couldn't allocate PD\n", __LINE__, __func__);
      goto err_cleanup;
   }

   // alloc and register buffers...

   commContext->commCfg = *commCfg;

   commContext->recvBufs = (char**)calloc(1, commCfg->bufNum * sizeof(char*) );
   commContext->recvMRs = (struct ibv_mr**)calloc(1, commCfg->bufNum * sizeof(struct ibv_mr*) );

   for(i=0; i < commCfg->bufNum; i++)
   {
      commContext->recvBufs[i] = __IBVSocket_allocAndRegisterBuf(
            commContext, commCfg->bufSize, &commContext->recvMRs[i] );
      if(!commContext->recvBufs[i])
      {
         fprintf(stderr, "%d:%s: Couldn't prepare recvBuf #%d\n", __LINE__, __func__, i+1);
         goto err_cleanup;
      }
   }

   commContext->sendBufs = (char**)calloc(1, commCfg->bufNum * sizeof(char*) );
   commContext->sendMRs = (struct ibv_mr**)calloc(1, commCfg->bufNum * sizeof(struct ibv_mr*) );

   for(i=0; i < commCfg->bufNum; i++)
   {
      commContext->sendBufs[i] = __IBVSocket_allocAndRegisterBuf(
         commContext, commCfg->bufSize, &commContext->sendMRs[i]);
      if(!commContext->sendBufs[i])
      {
         fprintf(stderr, "%d:%s: Couldn't prepare sendBuf #%d\n", __LINE__, __func__, i+1);
         goto err_cleanup;
      }
   }

   registerControlRes = __IBVSocket_registerBuf(
      commContext, (char*)&commContext->numUsedSendBufs,
      sizeof(commContext->numUsedSendBufs), &commContext->controlMR);
   if(registerControlRes < 0)
   {
      fprintf(stderr, "%d:%s: Couldn't register control memory region\n", __LINE__, __func__);
      goto err_cleanup;
   }

   registerControlResReset = __IBVSocket_registerBuf(
      commContext, (char*)&commContext->numUsedSendBufsReset,
      sizeof(commContext->numUsedSendBufsReset), &commContext->controlResetMR);
   if(registerControlRes < 0)
   {
      fprintf(stderr, "%d:%s: Couldn't register control memory region\n", __LINE__, __func__);
      goto err_cleanup;
   }

   // create completion channel and queues...

   commContext->recvCompChannel = ibv_create_comp_channel(commContext->context);
   if(!commContext->recvCompChannel)
   {
      fprintf(stderr, "%d:%s: Couldn't create comp channel\n", __LINE__, __func__);
      goto err_cleanup;
   }

   commContext->recvCQ = ibv_create_cq(
      commContext->context, commCfg->bufNum, commContext, commContext->recvCompChannel, 0);
   if(!commContext->recvCQ)
   {
      fprintf(stderr, "%d:%s: Couldn't create recv CQ\n", __LINE__, __func__);
      goto err_cleanup;
   }

   // note: 1+commCfg->bufNum here for the RDMA write usedBufs reset work (=> flow/flood control)
   commContext->sendCQ = ibv_create_cq(
      commContext->context, 1+commCfg->bufNum, NULL, NULL, 0);
   if(!commContext->sendCQ)
   {
      fprintf(stderr, "%d:%s: Couldn't create send CQ\n", __LINE__, __func__);
      goto err_cleanup;
   }

   // note: 1+commCfg->bufNum here for the RDMA write usedBufs reset work
   memset(&qpInitAttr, 0, sizeof(qpInitAttr) );

   qpInitAttr.send_cq = commContext->sendCQ;
   qpInitAttr.recv_cq = commContext->recvCQ;
   qpInitAttr.qp_type = IBV_QPT_RC;
   qpInitAttr.sq_sig_all = 1;
   qpInitAttr.cap.max_send_wr = 1+commCfg->bufNum;
   qpInitAttr.cap.max_recv_wr = commCfg->bufNum;
   qpInitAttr.cap.max_send_sge = 1;
   qpInitAttr.cap.max_recv_sge = 1;
   qpInitAttr.cap.max_inline_data = 0;

   if(rdma_create_qp(cm_id, commContext->pd, &qpInitAttr) )
   {
      fprintf(stderr, "%d:%s: Couldn't create QP\n", __LINE__, __func__);
      goto err_cleanup;
   }

   commContext->qp = cm_id->qp;

   // post initial recv buffers...

   for(i=0; i < commCfg->bufNum; i++)
   {
      __IBVSocket_postRecv(_this, commContext, i);
   }

   // prepare event notification...

   // initial event notification request
   if(ibv_req_notify_cq(commContext->recvCQ, 0) )
   {
      fprintf(stderr, "%s: Couldn't request CQ notification\n", __func__);
      goto err_cleanup;
   }

   *outCommContext = commContext;
   return true;


   //  error handling

err_cleanup:
   __IBVSocket_cleanupCommContext(commContext);

   *outCommContext = NULL;
   return false;
}

void __IBVSocket_cleanupCommContext(IBVCommContext* commContext)
{
   unsigned i;

   if(!commContext)
      return;

   if(commContext->recvCQ && commContext->numUnackedRecvCompChannelEvents)
      ibv_ack_cq_events(commContext->recvCQ, commContext->numUnackedRecvCompChannelEvents);

   if(commContext->qp)
   {
      if(ibv_destroy_qp(commContext->qp) )
         fprintf(stderr, "Couldn't destroy QP\n");
   }

   if(commContext->controlMR)
   {
      if(ibv_dereg_mr(commContext->controlMR) )
         fprintf(stderr, "%s: Failed to deregister controlMR\n", __func__);
   }

   if(commContext->controlResetMR)
   {
      if(ibv_dereg_mr(commContext->controlResetMR) )
         fprintf(stderr, "%s: Failed to deregister controlResetMR\n", __func__);
   }

   for(i=0; i < commContext->commCfg.bufNum; i++)
   {
      if(commContext->recvMRs && commContext->recvMRs[i] )
      {
         if(ibv_dereg_mr(commContext->recvMRs[i] ) )
            fprintf(stderr, "%s: Failed to deregister recvMR[%d]\n", __func__, i);
      }

      if(commContext->recvBufs)
         SAFE_FREE(commContext->recvBufs[i] );

      if(commContext->sendMRs && commContext->sendMRs[i] )
      {
         if(ibv_dereg_mr(commContext->sendMRs[i] ) )
            fprintf(stderr, "%s: Failed to deregister sendMR[%d]\n", __func__, i);
      }

      if(commContext->sendBufs)
         SAFE_FREE(commContext->sendBufs[i] );
   }

   SAFE_FREE(commContext->recvBufs);
   SAFE_FREE(commContext->sendBufs);
   SAFE_FREE(commContext->recvMRs);
   SAFE_FREE(commContext->sendMRs);

   free(commContext);
}

/**
 * Initializes a (local) IBVCommDest.
 */
void __IBVSocket_initCommDest(IBVCommContext* commContext, IBVCommDest* outDest)
{
   memcpy(outDest->verificationStr, IBVSOCKET_PRIVATEDATA_STR, IBVSOCKET_PRIVATEDATA_STR_LEN);

   //outDest->protocolVersion = IBVSOCKET_PRIVATEDATA_PROTOCOL_VER;
   Serialization::serializeUInt64( (char*)&outDest->protocolVersion,
      IBVSOCKET_PRIVATEDATA_PROTOCOL_VER);

   //outDest->rkey = commContext->controlMR->rkey;
   Serialization::serializeUInt( (char*)&outDest->rkey, commContext->controlMR->rkey);

   //outDest->vaddr = (uintptr_t)&commContext->numUsedSendBufs;
   Serialization::serializeUInt64( (char*)&outDest->vaddr,
      (uintptr_t)&commContext->numUsedSendBufs);

   //outDest->recvBufNum = commContext->commCfg.bufNum;
   Serialization::serializeUInt( (char*)&outDest->recvBufNum, commContext->commCfg.bufNum);

   //outDest->recvBufSize = commContext->commCfg.bufSize;
   Serialization::serializeUInt( (char*)&outDest->recvBufSize, commContext->commCfg.bufSize);
}

/**
 * Checks and parses a (remote) IBVCommDest.
 *
 * @param buf should usually be the private_data of the connection handshake
 * @param outDest will be alloced (if true is returned) and needs to be free'd by the caller
 * @return true if data is okay, false otherwise
 */
bool __IBVSocket_parseCommDest(const void* buf, size_t bufLen, IBVCommDest** outDest)
{
   IBVCommDest* dest = NULL;

   *outDest = NULL;


   // Note: "bufLen < ..." (and not "!="), because there might be some extra padding
   if(!buf || (bufLen < sizeof(*dest) ) )
   {
      fprintf(stderr, "%d:%s: bad private data. ptr: %p len: %d.\n",
         __LINE__, __func__, buf, (int)bufLen);

      return false;
   }

   dest = (IBVCommDest*)malloc(sizeof(*dest) );
   if(!dest)
      return false;

   memcpy(dest, buf, sizeof(*dest) );

   if(memcmp(dest->verificationStr, IBVSOCKET_PRIVATEDATA_STR, IBVSOCKET_PRIVATEDATA_STR_LEN) )
      goto err_cleanup;

   uint64_t destProtoVer;
   unsigned destProtoLen;
   Serialization::deserializeUInt64(
      (const char*)&dest->protocolVersion, 8L, &destProtoVer, &destProtoLen);

   dest->protocolVersion = destProtoVer; // replace with native-endian version

   if(destProtoVer != IBVSOCKET_PRIVATEDATA_PROTOCOL_VER)
      goto err_cleanup;


   unsigned destRKey;
   unsigned destRKeyLen;
   Serialization::deserializeUInt(
      (const char*)&dest->rkey, 8L, &destRKey, &destRKeyLen);

   dest->rkey = destRKey; // replace with native-endian version

   uint64_t destVAddr;
   unsigned destVAddrLen;
   Serialization::deserializeUInt64(
      (const char*)&dest->vaddr, 8L, &destVAddr, &destVAddrLen);

   dest->vaddr = destVAddr; // replace with native-endian version

   unsigned destRecvBufNum;
   unsigned destRecvBufNumLen;
   Serialization::deserializeUInt(
      (const char*)&dest->recvBufNum, 8L, &destRecvBufNum, &destRecvBufNumLen);

   dest->recvBufNum = destRecvBufNum; // replace with native-endian version

   unsigned destRecvBufSize;
   unsigned destRecvBufSizeLen;
   Serialization::deserializeUInt(
      (const char*)&dest->recvBufSize, 8L, &destRecvBufSize, &destRecvBufSizeLen);

   dest->recvBufSize = destRecvBufSize; // replace with native-endian version


   *outDest = dest;

   return true;


err_cleanup:
   SAFE_FREE(dest);

   return false;
}


/**
 * @param commContext passed seperately because it's not the _this->commContext during
 *    accept() of incoming connections
 * @return 0 on success, -1 on error
 */
int __IBVSocket_postRecv(IBVSocket* _this, IBVCommContext* commContext, size_t bufIndex)
{
   struct ibv_sge list;
   struct ibv_recv_wr wr;
   struct ibv_recv_wr* bad_wr;
   int postRes;

   list.addr = (uint64_t)commContext->recvBufs[bufIndex];
   list.length = commContext->commCfg.bufSize;
   list.lkey = commContext->recvMRs[bufIndex]->lkey;

   wr.next = NULL;
   wr.wr_id = bufIndex + IBVSOCKET_RECV_WORK_ID_OFFSET;
   wr.sg_list = &list;
   wr.num_sge = 1;

   postRes = ibv_post_recv(commContext->qp, &wr, &bad_wr);
   if(unlikely(postRes) )
   {
      fprintf(stderr, "%d:%s: ibv_post_recv failed. ErrCode: %d\n", __LINE__, __func__, postRes);

      return -1;
   }

   return 0;
}

/**
 * @return 0 on success, -1 on error
 */
int __IBVSocket_postWrite(IBVSocket* _this, IBVCommDest* remoteDest,
   struct ibv_mr* localMR, char* localBuf, int bufLen)
{
   IBVCommContext* commContext = _this->commContext;
   struct ibv_sge list;
   struct ibv_send_wr wr;
   struct ibv_send_wr *bad_wr;
   int postRes;
   int waitRes;

   list.addr = (uint64_t)localBuf;
   list.length = bufLen;
   list.lkey = localMR->lkey;

   wr.wr.rdma.remote_addr = remoteDest->vaddr;
   wr.wr.rdma.rkey = remoteDest->rkey;

   wr.wr_id      = IBVSOCKET_WRITE_WORK_ID;
   wr.sg_list    = &list;
   wr.num_sge    = 1;
   wr.opcode     = IBV_WR_RDMA_WRITE;
   wr.send_flags = IBV_SEND_SIGNALED;
   wr.next       = NULL;

   postRes = ibv_post_send(commContext->qp, &wr, &bad_wr);
   if(unlikely(postRes) )
   {
      fprintf(stderr, "%d:%s: ibv_post_send() failed: %s\n", __LINE__, __func__,
         strerror(-postRes) );

      return -1;
   }

   waitRes = __IBVSocket_waitForTotalSendCompletion(_this,
      commContext->incompleteSend.numAvailable, 1, 0);
   if(unlikely(waitRes) )
      return -1;

   commContext->incompleteSend.numAvailable = 0;

   return 0;
}

/**
 * @return 0 on success, -1 on error
 */
int __IBVSocket_postRead(IBVSocket* _this, IBVCommDest* remoteDest,
   struct ibv_mr* localMR, char* localBuf, int bufLen)
{
   IBVCommContext* commContext = _this->commContext;
   struct ibv_sge list;
   struct ibv_send_wr wr;
   struct ibv_send_wr *bad_wr;
   int postRes;
   int waitRes;

   list.addr = (uint64_t) localBuf;
   list.length = bufLen;
   list.lkey = localMR->lkey;

   wr.wr.rdma.remote_addr = remoteDest->vaddr;
   wr.wr.rdma.rkey = remoteDest->rkey;

   wr.wr_id      = IBVSOCKET_READ_WORK_ID;
   wr.sg_list    = &list;
   wr.num_sge    = 1;
   wr.opcode     = IBV_WR_RDMA_READ;
   wr.send_flags = IBV_SEND_SIGNALED;
   wr.next       = NULL;

   postRes = ibv_post_send(commContext->qp, &wr, &bad_wr);
   if(unlikely(postRes) )
   {
      fprintf(stderr, "%d:%s: ibv_post_send() failed: %s\n", __LINE__, __func__,
         strerror(-postRes) );

      return -1;
   }

   waitRes = __IBVSocket_waitForTotalSendCompletion(_this,
      commContext->incompleteSend.numAvailable, 0, 1);
   if(unlikely(waitRes) )
      return -1;

   commContext->incompleteSend.numAvailable = 0;

   return 0;
}

/**
 * @return 0 on success, -1 on error
 */
int __IBVSocket_postSend(IBVSocket* _this, size_t bufIndex, int bufLen)
{
   IBVCommContext* commContext = _this->commContext;
   struct ibv_sge list;
   struct ibv_send_wr wr;
   struct ibv_send_wr *bad_wr;
   int postRes;

   list.addr   = (uint64_t)commContext->sendBufs[bufIndex];
   list.length = bufLen;
   list.lkey   = commContext->sendMRs[bufIndex]->lkey;

   wr.wr_id      = bufIndex + IBVSOCKET_SEND_WORK_ID_OFFSET;
   wr.next       = NULL;
   wr.sg_list    = &list;
   wr.num_sge    = 1;
   wr.opcode     = IBV_WR_SEND;
   wr.send_flags = IBV_SEND_SIGNALED;

   postRes = ibv_post_send(commContext->qp, &wr, &bad_wr);
   if(unlikely(postRes) )
   {
      fprintf(stderr, "%d:%s: ibv_post_send() failed: %s\n", __LINE__, __func__,
         strerror(-postRes) );

      return -1;
   }

   return 0;
}


/**
 * @return 1 on success, 0 on timeout, -1 on error
 */
int __IBVSocket_recvWC(IBVSocket* _this, int timeoutMS, struct ibv_wc* outWC)
{
   IBVCommContext* commContext = _this->commContext;
   int numEvents;
   size_t bufIndex;
   int waitRes;

   do
   {
      numEvents = ibv_poll_cq(commContext->recvCQ, 1, outWC);
      if(unlikely(numEvents < 0) )
      {
         fprintf(stderr, "%s: poll CQ failed %d\n", __func__, numEvents);
         return -1;
      }

      if(!numEvents)
      {
         waitRes = __IBVSocket_waitForRecvCompletionEvent(_this, timeoutMS);
         if(waitRes > 0)
            continue;
         else
            return waitRes;
      }

   } while(!numEvents);

   if(unlikely(outWC->status != IBV_WC_SUCCESS) )
   {
      fprintf(stderr, "%s: Failed status %d for wr_id %d\n",
         __func__, outWC->status, (int) outWC->wr_id);
      return -1;
   }

   bufIndex = outWC->wr_id - IBVSOCKET_RECV_WORK_ID_OFFSET;

   if(unlikely(bufIndex >= commContext->commCfg.bufNum) )
   {
      fprintf(stderr, "%s: Completion for unknown/invalid wr_id %d\n",
         __func__, (int)outWC->wr_id);
      return -1;
   }

   // receive completed
   //printf("%s: Recveived %u bytes.\n", __func__, outWC->byte_len); // debug in

   return 1;
}


/**
 * @return 1 on available data, 0 on timeout, -1 on error
 */
int __IBVSocket_waitForRecvCompletionEvent(IBVSocket* _this, int timeoutMS)
{
   // Note: This will also be called with timeoutMS==0 from nonblockingRecvCheck
   //    (just be aware of that).

   enum {COMPLETION_FD = 0, ASYNC_EVENT_FD = 1};

   IBVCommContext* commContext = _this->commContext;
   struct ibv_cq*  ev_cq;
   void*           ev_ctx;
   struct pollfd pollFDs[1/*2*/];
   int pollRes;
   int checkRes;

   pollFDs[COMPLETION_FD].fd = commContext->recvCompChannel->fd;
   pollFDs[COMPLETION_FD].events = POLLIN;
   pollFDs[COMPLETION_FD].revents = 0;

   // note: problem: accepted sockets don't have their own async event channel
//   pollFDs[ASYNC_EVENT_FD].fd = commContext->context->async_fd;
//   pollFDs[ASYNC_EVENT_FD].events = POLLIN;
//   pollFDs[ASYNC_EVENT_FD].revents = 0;

   do
   {
      // Note: We use pollTimeoutMS to check the conn every few secs (otherwise we might
      //    wait for a very long time in case the other side disconnected silently)
      int pollTimeoutMS = FHGFS_MIN(10000, timeoutMS);

      pollRes = poll(pollFDs, 1/*2*/, pollTimeoutMS);
      if(pollRes < 0)
      {
         fprintf(stderr, "%s: poll() error: %d", __func__, pollRes);
         return -1;
      }

      if(pollRes == 0)
      {

         //fprintf(stderr, "%s: poll() timed out\n", __func__); // debug in

         // Note: we check "timeoutMS != 0" here because we don't want to run the
         //    connCheck each time this method is called from nonblockingRecvCheck
         if(timeoutMS)
         {
            checkRes = IBVSocket_checkConnection(_this);
            if(checkRes < 0)
               return -1;
         }

         timeoutMS -= pollTimeoutMS;
         if(!timeoutMS)
            return 0;

         continue;
      }

      if(pollFDs[COMPLETION_FD].revents)
      {
         //fprintf(stderr, "%s: poll() received a completion event\n", __func__); // debug in
      }

//      if(pollFDs[ASYNC_EVENT_FD].revents)
//      {
//         fprintf(stderr, "%s: poll() received an async event\n", __func__);
//         return -1;
//      }

   } while(!pollRes);

   //printf("%s:%d: waiting for event notification\n", __func__, __LINE__); // debug in

   if(ibv_get_cq_event(commContext->recvCompChannel, &ev_cq, &ev_ctx) )
   {
      fprintf(stderr, "%s: Failed to get cq_event\n", __func__);
      return -1;
   }

   //printf("%s:%d: event notification received\n", __func__, __LINE__); // debug in

   if(unlikely(ev_cq != commContext->recvCQ) )
   {
      fprintf(stderr, "%s: CQ event for unknown CQ %p\n", __func__, ev_cq);
      return -1;
   }

   //printf("%s:%d: requesting next event notification\n", __func__, __LINE__); // debug in

   if(ibv_req_notify_cq(commContext->recvCQ, 0) )
   {
      fprintf(stderr, "%s: Couldn't request CQ notification\n", __func__);
      return -1;
   }


   // ack is expensive, so we gather and ack multiple events
   // note: spec says we need this, but current send_bw.c & co don't use ibv_ack_cq_events.

   commContext->numUnackedRecvCompChannelEvents++;
   if(commContext->numUnackedRecvCompChannelEvents == IBVSOCKET_EVENTS_GATHER_NUM)
   { // ack events and reset counter
      ibv_ack_cq_events(commContext->recvCQ, commContext->numUnackedRecvCompChannelEvents);
      commContext->numUnackedRecvCompChannelEvents = 0;
   }

   return 1;
}


/**
 * @return number of completed elements or -1 in case of an error
 */
int __IBVSocket_waitForTotalSendCompletion(IBVSocket* _this,
   int numSendElements, int numWriteElements, int numReadElements)
{
   IBVCommContext* commContext = _this->commContext;
   int numElements;
   int i;
   size_t bufIndex;
   struct ibv_wc wc[2];


   do
   {
      numElements = ibv_poll_cq(commContext->sendCQ, 2, wc);
      if(unlikely(numElements < 0) )
      {
         fprintf(stderr, "%d:%s: bad ibv_poll_cq result: %d\n", __LINE__, __func__,
            numElements);

         return -1;
      }

      // for each completion element
      for(i=0; i < numElements; i++)
      {
         if(unlikely(wc[i].status != IBV_WC_SUCCESS) )
         {
            fprintf(stderr, "%d:%s: bad wc status %d\n", __LINE__, __func__,
               wc[i].status);

            return -1;
         }

         switch(wc[i].opcode)
         {
            case IBV_WC_SEND:
            {
               bufIndex = wc[i].wr_id - IBVSOCKET_SEND_WORK_ID_OFFSET;

               if(unlikely(bufIndex >= commContext->commCfg.bufNum) )
               {
                  fprintf(stderr, "%d:%s: bad send completion wr_id 0x%x\n", __LINE__, __func__,
                     (int)wc[i].wr_id);

                  return -1;
               }

               if(likely(numSendElements) )
                  numSendElements--;
               else
               {
                  fprintf(stderr, "%d:%s: received bad/unexpected send completion\n",
                     __LINE__, __func__);

                  return -1;
               }

            } break;

            case IBV_WC_RDMA_WRITE:
            {
               if(unlikely(wc[i].wr_id != IBVSOCKET_WRITE_WORK_ID) )
               {
                  fprintf(stderr, "%d:%s: bad write completion wr_id 0x%x\n", __LINE__, __func__,
                     (int)wc[i].wr_id);

                  return -1;
               }

               if(likely(numWriteElements) )
                  numWriteElements--;
               else
               {
                  fprintf(stderr, "%d:%s: received bad/unexpected RDMA write completion\n",
                     __LINE__, __func__);

                  return -1;
               }
            } break;

            case IBV_WC_RDMA_READ:
            {
               if(unlikely(wc[i].wr_id != IBVSOCKET_READ_WORK_ID) )
               {
                  fprintf(stderr, "%d:%s: bad read completion wr_id 0x%x\n", __LINE__, __func__,
                     (int)wc[i].wr_id);

                  return -1;
               }

               if(likely(numReadElements) )
                  numReadElements--;
               else
               {
                  fprintf(stderr, "%d:%s: received bad/unexpected RDMA read completion\n",
                     __LINE__, __func__);

                  return -1;
               }
            } break;

            default:
            {
               fprintf(stderr, "%d:%s: bad/unexpected completion opcode %d\n", __LINE__,
                  __func__, wc[i].opcode);

               return -1;
            } break;

         } // end of switch

      } // end of for-loop

   } while(numSendElements || numWriteElements || numReadElements);

   return 0;
}

/**
 * @return 0 on success, -1 on error (e.g. timeout)
 */
int __IBVSocket_waitForUsedSendBufsReset(IBVSocket* _this)
{
   IBVCommContext* commContext = _this->commContext;

   if(commContext->numUsedSendBufs == commContext->commCfg.bufNum)
      commContext->numUsedSendBufs = 0;

   return 0;
}


/**
 * @return 0 on success, -1 on error
 */
int IBVSocket_checkConnection(IBVSocket* _this)
{
   int postRes;
   IBVCommContext* commContext = _this->commContext;

   // note: we read a remote value into the numUsedSendBufsReset field, which is actually
   //    meant for something else, so we need to reset the value afterwards

   //printf("%d:%s: post rdma_read to check connection...\n", __LINE__, __func__); // debug in

   postRes = __IBVSocket_postRead(_this, _this->remoteDest, commContext->controlResetMR,
      (char*)&commContext->numUsedSendBufsReset, sizeof(commContext->numUsedSendBufsReset) );
   if(postRes)
      return -1;

   commContext->numUsedSendBufsReset = 0;

   //printf("%d:%s: rdma_read succeeded\n", __LINE__, __func__); // debug in

   return 0;
}

/**
 * @return <0 on error, 0 if recv would block, >0 if recv would not block
 */
ssize_t IBVSocket_nonblockingRecvCheck(IBVSocket* _this)
{
   IBVCommContext* commContext = _this->commContext;
   struct ibv_wc* wc = &commContext->incompleteRecv.wc;
   int completionRes;
   int numEvents;
   size_t bufIndex;

   // check for inclomplete recv
   if(commContext->incompleteRecv.isAvailable)
   {
      int completedOffset = commContext->incompleteRecv.completedOffset;
      size_t availableLen = commContext->incompleteRecv.wc.byte_len - completedOffset;

      return availableLen;
   }

   // remove event channel notification (to avoid endless false alerts)
   completionRes = __IBVSocket_waitForRecvCompletionEvent(_this, 0);
   if(completionRes < 0)
      goto err_invalidateSock;


   // recv one packet (if available) and add it as incompleteRecv
   numEvents = ibv_poll_cq(_this->commContext->recvCQ, 1, wc);
   if(numEvents < 0)
   {
      fprintf(stderr, "%s: poll CQ failed %d\n", __func__, numEvents);
      goto err_invalidateSock;
   }
   if(!numEvents)
      return 0;

   if(wc->status != IBV_WC_SUCCESS)
   {
      fprintf(stderr, "%s: Failed status %d for wr_id %d\n",
         __func__, wc->status, (int) wc->wr_id);
      goto err_invalidateSock;
   }

   bufIndex = wc->wr_id - IBVSOCKET_RECV_WORK_ID_OFFSET;

   if(bufIndex >= commContext->commCfg.bufNum)
   {
      fprintf(stderr, "%s: Completion for unknown/invalid wr_id %d\n", __func__, (int)wc->wr_id);
      goto err_invalidateSock;
   }

   // prepare to continue later
   bufIndex = wc->wr_id - IBVSOCKET_RECV_WORK_ID_OFFSET;

   commContext->incompleteRecv.completedOffset = 0;
   commContext->incompleteRecv.isAvailable = 1;

   return wc->byte_len;

err_invalidateSock:
   _this->errState = -1;
   return -1;
}

/**
 * Call this after accept() to find out whether more events are waiting (for which
 * no notification would not be delivered through the file descriptor).
 *
 * @return true if more events are waiting and accept() should be called again
 */
bool IBVSocket_checkDelayedEvents(IBVSocket* _this)
{
   bool retVal = false;
   struct rdma_cm_event* event;

   // check for events in the delay queue
   if(_this->delayedCmEventsQ->size() > 0)
      return true;


   // Switch channel fd to non-blocking, check for waiting events and switch back to blocking.
   // (Quite inefficient, but we really don't have to care about efficiency in this case.)

   // Note: We do this to avoid race conditions (lost events) before we're waiting for
   // new notifications with poll()

   // change mode of the connection manager channel to non-blocking
   int oldChannelFlags = fcntl(IBVSocket_getConnManagerFD(_this), F_GETFL);

   int setNewFlagsRes = fcntl(
      IBVSocket_getConnManagerFD(_this), F_SETFL, oldChannelFlags | O_NONBLOCK);
   if(setNewFlagsRes < 0)
   {
      fprintf(stderr, "%d:%s: set conn manager channel non-blocking failed. errno: %d\n",
         __LINE__, __func__, errno);
      return false;
   }

   // (non-blocking) check for new events
   if(rdma_get_cm_event(_this->cm_channel, &event) )
   {
      // non-blocking mode, so we ignore "pseudo-errors" here
   }
   else
   { // incoming event available
      //printf("%d:%s: enqueueing an event (during non-blocking check): %d (%s)\n",
      //   __LINE__, __func__, event->event, rdma_event_str(event->event) ); // debug in

      _this->delayedCmEventsQ->push(event);

      retVal = true;
   }


   // change channel mode back to blocking
   int setOldFlagsRes = fcntl(IBVSocket_getConnManagerFD(_this), F_SETFL, oldChannelFlags);
   if(setOldFlagsRes < 0)
   {
      fprintf(stderr, "%d:%s: set conn manager channel blocking failed. errno: %d\n",
         __LINE__, __func__, errno);
      return false;
   }


   return retVal;
}

/**
 * Note: This is provided simply because it exists, but somehow the connection tear-down process
 *    often works better (without delay) if this explicit disconnect is avoided. In fact, it seems
 *    a receiver might sometimes even hang infinitely if this is used.
 * Note2: The above note is probably no longer true since the other end has a better
 *    handling of disconnect events (todo: check that)
 */
void __IBVSocket_disconnect(IBVSocket* _this, bool waitForEvent)
{
   struct rdma_cm_event *event;
   int disconnectRes;

   disconnectRes = rdma_disconnect(_this->cm_id);
   if(disconnectRes)
   {
      perror("rdma_disconnect");
      fprintf(stderr, "%d:%s: rdma disconnect error\n", __LINE__, __func__);

      return;
   }

   if(!_this->cm_channel || !waitForEvent)
      return;

   rdma_get_cm_event(_this->cm_channel, &event);
   if(event->event != RDMA_CM_EVENT_DISCONNECTED)
   {
      fprintf(stderr, "%s: unexpected event during disconnect %d: %s\n",
         __func__, event->event, rdma_event_str(event->event) );
   }

   rdma_ack_cm_event(event);
}

void __IBVSocket_close(IBVSocket* _this)
{
   SAFE_FREE(_this->remoteDest);

   if(_this->delayedCmEventsQ)
   { // ack all queued events
      while(_this->delayedCmEventsQ->size() )
      {
         struct rdma_cm_event* nextEvent =  _this->delayedCmEventsQ->front();

         rdma_ack_cm_event(nextEvent);

         _this->delayedCmEventsQ->pop();
      }

      delete(_this->delayedCmEventsQ);
   }

   if(_this->commContext)
      __IBVSocket_cleanupCommContext(_this->commContext);

   if(_this->cm_id)
      rdma_destroy_id(_this->cm_id);
   if(_this->cm_channel)
      rdma_destroy_event_channel(_this->cm_channel);
}

bool IBVSocket_getSockValid(IBVSocket* _this)
{
   return _this->sockValid;
}

int IBVSocket_getRecvCompletionFD(IBVSocket* _this)
{
   IBVCommContext* commContext = _this->commContext;

   return commContext ? commContext->recvCompChannel->fd : (-1);
}

int IBVSocket_getConnManagerFD(IBVSocket* _this)
{
   return _this->cm_channel ? _this->cm_channel->fd : (-1);
}

#endif // FHGFS_OPENTK_IBVERBS

